### File name: quadratic.approx.R

# packages needed: distr, use the following to install
# install.packages("distr")

# constants used
MIN = 1e-20
library(distr)
library(adapt)
distroptions(DefaultNrFFTGridPointsExponent = 15)

# given the mean mu, variance matrix sigma, similarity matrix A and the quantity d,
# find the density of random variable Y = X^T A X using 2-cum or 4-cum chi-square approximation
dchiapprox = function(mu, sigma, A, y, method=4, pos=0)
{
   if (sum(abs(mu)) != 0) method = 4 # only 4-cum approximation is valid
   if (method == 2 || method == "2cum" || method == "2-cum")
   {
      temp = two_cum_chi (sigma, A, pos=pos)
      df1 = temp$dfpos
      alpha1 = temp$alphapos
      df2 = temp$dfneg
      alpha2 = temp$alphaneg
      if (df2 == 0)
      {
         if (df1 == 0) 
         {
            printf ("Warning: the statistic is constant 0")
            break
         }
         fy = dchisq(y*alpha1,df1,ncp=0)*alpha1
      } else
      {
         if (df1 == 0) 
         {
            fy = dchisq(-y*alpha2,df2,ncp=0)*alpha2
         } else
         {
            require(distr)
            ch1 = Chisq(df=df1, ncp=0)
            ch2 = Chisq(df=df2, ncp=0)
            twchi.dist = ch1/alpha1 - ch2/alpha2
            fy = d(twchi.dist)(y)
         }
      }
   } else if (method == 4 || method == "4cum" || method == "4-cum")
   {
      temp = four_cum_chi (mu,sigma, A, pos=pos)
      df1 = temp$dfpos
      alpha1 = temp$alphapos
      beta1 = temp$betapos
      delta1 = temp$deltapos
      df2 = temp$dfneg
      alpha2 = temp$alphaneg
      beta2 = temp$betaneg
      delta2 = temp$deltaneg
      shift = temp$shift
      if (df2 == 0)  
      {
         if (df1 == 0) 
         {
            printf ("Warning: the statistic is constant 0")
            break
         }
         fy = dchisq(y*alpha1+beta1,df1,ncp=delta1)*alpha1
      } else 
      {
         if (df1 == 0)
         {
            fy = dchisq(-y*alpha2+beta2,df2,ncp=delta2)*alpha2
         } else
         {
            require(distr)
            ch1 = Chisq(df=df1, ncp=delta1) - beta1
            ch2 = Chisq(df=df2, ncp=delta2) - beta2
            twchi.dist = ch1/alpha1 - ch2/alpha2 + shift
            fy = d(twchi.dist)(y)
         }
      }
   } else 
   {
      printf ("Please choose the appropriate method for approximation!")
   }
   fy
}

# given the mean mu, variance matrix sigma, similarity matrix A and the quantity d,
# find the probability P{X^T A X <= d} using 2-cum or 4-cum chi-square approximation
pchiapprox = function(mu, sigma, A, d, method=4, pos=0)
{
   if (sum(abs(mu)) != 0) method = 4 # only 4-cum approximation is valid
   if (method == 2 || method == "2cum" || method == "2-cum")
   {
      temp = two_cum_chi (sigma, A, pos=pos)
      df1 = temp$dfpos
      alpha1 = temp$alphapos
      df2 = temp$dfneg
      alpha2 = temp$alphaneg
      if (df2 == 0)
      {
         if (df1 == 0) 
         {
            printf ("Warning: the statistic is constant 0")
            break
         }
         prob = pchisq(d*alpha1,df1,ncp=0)
      } else
      {
         if (df1 == 0) 
         {
            prob = 1 - pchisq(-d*alpha2,df2,ncp=0)
         } else
         {
            require(distr)
            ch1 = Chisq(df=df1, ncp=0)
            ch2 = Chisq(df=df2, ncp=0)
            twchi.dist = ch1/alpha1 - ch2/alpha2
            prob = p(twchi.dist)(d)
         }
      }
   } else if (method == 4 || method == "4cum" || method == "4-cum")
   {
      temp = four_cum_chi (mu,sigma, A, pos=pos)
      df1 = temp$dfpos
      alpha1 = temp$alphapos
      beta1 = temp$betapos
      delta1 = temp$deltapos
      df2 = temp$dfneg
      alpha2 = temp$alphaneg
      beta2 = temp$betaneg
      delta2 = temp$deltaneg
      shift = temp$shift
      if (df2 == 0)  
      {
         if (df1 == 0) 
         {
            printf ("Warning: the statistic is constant 0")
            break
         }
         prob = pchisq(d*alpha1+beta1,df1,ncp=delta1)
      } else 
      {
         if (df1 == 0)
         {
            prob = 1 - pchisq(-d*alpha2+beta2,df2,ncp=delta2)
         } else
         {
            require(distr)
            ch1 = Chisq(df=df1, ncp=delta1) - beta1
            ch2 = Chisq(df=df2, ncp=delta2) - beta2
            twchi.dist = ch1/alpha1 - ch2/alpha2 + shift
            prob = p(twchi.dist)(d)
         }
      }
   } else 
   {
      printf ("Please choose the appropriate method for approximation!")
   }
   prob
}

# given the mean mu, variance matrix sigma, similarity matrix A and the significance level alpha,
# find the quantile dstar such that alpha=P{X^T A X <= dstar} using 2-cum or 4-cum chi-square approximation
qchiapprox = function(mu, sigma, A, alpha, method=4, pos=0)
{
   if (sum(abs(mu)) != 0) method = 4 # only 4-cum approximation is valid
   if (method == 2 || method == "2cum" || method == "2-cum")
   {
      temp = two_cum_chi (sigma, A, pos=pos)
      df1 = temp$dfpos
      alpha1 = temp$alphapos
      df2 = temp$dfneg
      alpha2 = temp$alphaneg
      if (df2 == 0)
      {
         if (df1 == 0) {
            printf ("Warning: the statistic is constant 0")
            break
         }
         dstar = qchisq(alpha,df1,ncp=0)/alpha1
      } else
      {
         if (df1 == 0) 
         {
            dstar = -qchisq(1-alpha,df2,ncp=0)/alpha2
         } else
         {
            require(distr)
            ch1 = Chisq(df=df1, ncp=0)
            ch2 = Chisq(df=df2, ncp=0)
            twchi.dist = ch1/alpha1 - ch2/alpha2
            dstar = q(twchi.dist)(alpha)
         }
      }
   } else if (method == 4  || method == "4cum" || method == "4-cum")
   {
      temp = four_cum_chi (mu,sigma, A, pos=pos)
      df1 = temp$dfpos
      alpha1 = temp$alphapos
      beta1 = temp$betapos
      delta1 = temp$deltapos
      df2 = temp$dfneg
      alpha2 = temp$alphaneg
      beta2 = temp$betaneg
      delta2 = temp$deltaneg
      shift = temp$shift
      if (df2 == 0)  
      {
         if (df1 == 0)
         {
            printf ("Warning: the statistic is constant 0")
            break
         }
         dstar = (qchisq(alpha,df1,ncp=delta1)-beta1)/alpha1
      } else 
      {
         if (df1 == 0) 
         {
            dstar = -(qchisq(1-alpha,df2,ncp=delta2)-beta2)/alpha2
         } else
         {
            require(distr)
            ch1 = Chisq(df=df1, ncp=delta1) - beta1
            ch2 = Chisq(df=df2, ncp=delta2) - beta2
            twchi.dist = ch1/alpha1 - ch2/alpha2 + shift
            dstar = q(twchi.dist)(alpha)
         }
      }
   } else {
      printf ("Please choose the appropriate method for approximation!")
   }
   dstar
}

samplesize = function(p, q, A.alt, n, m, sigma.alt=NULL,
   r=NULL, A.null=NULL, sigma.null=NULL, pvalue=0.05, power=0.8, 
   method=4, tol=0.0001, maxite=100)
# p is the haplotype frequency for population 1
# q is the haplotype frequency for population 2
# A.alt is the similarity matrix under alternative hypothesis
# sigma.alt is the variance matrix under alternative hypothesis
# r is the haplotype frequency combining population 1 and population 2
# A.null is the similarity matrix under null hypothesis
# sigma.null is the variance matrix under null hypothesis
# pvalue is the required significance level
# power is the required power of detection
# method = 4: 4-cum approximation to get critical value for rejection
#        = 2: 2-cum approximation to get critical value for rejection
{
   nhap.alt = length(p)
   semipos.alt = 1
   temp = eigen(A.alt,symmetric=T)$values
   if (temp[nhap.alt] < -MIN) semipos.alt = 2

   if (is.null(sigma.alt)) sigma.alt = diag(p/n+q/m,nhap.alt)-p%*%t(p)/n-q%*%t(q)/m;  # sigma_s
   if (is.null(r)) r = (p*n + q*m)/(n+m)
   nhap.null = length(r)
   if (is.null(sigma.null)) sigma.null = (diag(r,nhap.null)-r%*%t(r))*(1/n+1/m)
   if (is.null(A.null))
   {
      A.null = A.alt
      semipos.null = semipos.alt
   } else
   {
      semipos.null = 1
      temp = eigen(A.null,symmetric=T)$values
      if (temp[nhap.null] < -MIN) semipos.null = 2
   }

   cc.low = 2/min(n,m)
   cc.up = 1000
   cc = cc.up
   kk = 0
   pow = power + 1
   while (abs(pow-power) > tol && kk < maxite)
   {
      # first find the critical value dstar at significance level pvalue under null hypothesis
      sigma = sigma.null/cc
      dstar = qchiapprox (0, sigma, A.null, 1 - pvalue, pos = semipos.null, method = method)

      # then find the power for this given dstar under alternative hypothesis
      sigma = sigma.alt/cc
      pow = 1 - pchiapprox (p-q, sigma, A.alt, dstar, method=method, pos=semipos.alt)

      if (pow < power && kk == 0)
      {
         print("Warning: you need sample size greater than the following")
         list (size.of.pop1 = n*cc, size.of.pop2 = m*cc, power = pow)
         break
      }
      if (pow < power)
      {
         cc.low = cc
      } else
      {
         cc.up = cc
      }
      cc = (cc.low+cc.up)/2
      kk = kk + 1
   }
   list (size.of.pop1 = n*cc, size.of.pop2 = m*cc, power = pow)
}

# require: variance matrix sigma, similarity matrix A
# options: pos=0, do not know if A is positive semi-definite or not
#          pos=1, A is positive semi-definite
#          pos=2 or other, A is not positive semi-definite
# return: dfpos(degrees of freedom), alphapos(scale),
#         dfneg(degrees of freedom), alphaneg(scale)
two_cum_chi = function(sigma, A, pos=0)
{
   nhap = nrow(A)
   if (pos == 0)
   {
      pos = 1
      temp = eigen(A,symmetric=T)$values
      if (temp[nhap] < -MIN) pos = 2
   }

   if (pos == 1)               # if A is positive semi-definite
   {
      zsigma = A%*%sigma
      temp1 = 0
      for(k in 1:nhap) temp1 = temp1 + zsigma[k,k]
      temp2 = 0
      for(k1 in 1:nhap)
         for(k2 in 1:nhap) temp2 = temp2 + zsigma[k1,k2]*zsigma[k2,k1]
      df = alpha = 0
      if (temp2 != 0)
      {
         df = temp1^2/temp2
         alpha = temp1/temp2
      }
      return(list(dfpos=df, alphapos=alpha, dfneg=0, alphaneg=0))

   } else                          # if A is not positive semi-definite
   {
      temp = get_weight_center_shift(0, sigma, A)
      rr = temp$rank
      lambda = temp$weight[1:rr]
      # center and shift are both 0 under null

      wpos = lambda[lambda > MIN]
      rpos = length(wpos)
      wneg = -lambda[lambda < -MIN]
      rneg = length(wneg)

      # use the difference of two 2-cumulants chi-square approx.
      df1 = alpha1 = df2 = alpha2 = 0
      if (rpos > 0)
      {
         trace = sum(wpos)
         temp2 = sum(wpos^2)
         df1 = trace^2/temp2
         alpha1 = trace/temp2
      }
      if (rneg > 0)
      {
         trace = sum(wneg)
         temp2 = sum(wneg^2)
         df2 = trace^2/temp2
         alpha2 = trace/temp2
      }
      return(list(dfpos=df1, alphapos=alpha1, dfneg=df2, alphaneg=alpha2))
   }
}

# require: mean mu, variance matrix sigma, similarity matrix A
# options: pos=0, do not know if A is positive semi-definite or not
#          pos=1, A is positive semi-definite
#          pos=2 or other, A is not positive semi-definite
# return: dfpos(degrees of freedom), deltapos(center), alphapos(scale), betapos(positive shift),
#         dfneg(degrees of freedom), deltaneg(center), alphaneg(scale), betaneg(negative shift),
#         and shift (overall shift)
four_cum_chi = function(mu,sigma,A,pos=0)
{
   nhap = nrow(A)
   if (pos == 0)
   {
      pos = 1
      temp = eigen(A,symmetric=T)$values
      if (temp[nhap] < -MIN) pos = 2
   }
   if (pos == 1)               # if A is positive semi-definite
   {
      temp = get_noncentral_chi(mu,sigma,A)
      return(list(dfpos=temp$df,deltapos=temp$delta,alphapos=temp$alpha,betapos=temp$beta,
         dfneg=0,deltaneg=0,alphaneg=0,betaneg=0,shift=0))

   } else
   {
      temp = get_weight_center_shift(mu, sigma, A)
      rr = temp$rank
      lambda = temp$weight[1:rr]
      center = temp$center[1:rr]
      shift = temp$shift

      wpos = lambda[lambda > MIN]
      rpos = length(wpos)
      cpos = center[lambda > MIN]
      wneg = -lambda[lambda < -MIN]
      rneg = length(wneg)
      cneg = center[lambda < -MIN]

      # use the difference of two 4-cumulants chi-square approx.
      alpha1 = beta1 = delta1 = df1 = 0
      alpha2 = beta2 = delta2 = df2 = 0
      if (rpos > 0)
      {
         temp = get_noncentral_chi(cpos, diag(1,rpos), diag(wpos,rpos))
         alpha1 = temp$alpha
         beta1 = temp$beta
         delta1 = temp$delta
         df1 = temp$df
      }
      if (rneg > 0)
      {
         temp = get_noncentral_chi(cneg, diag(1,rneg), diag(wneg,rneg))
         alpha2 = temp$alpha
         beta2 = temp$beta
         delta2 = temp$delta
         df2 = temp$df
      }
      return(list(dfpos=df1,deltapos=delta1,alphapos=alpha1,betapos=beta1,
         dfneg=df2,deltaneg=delta2,alphaneg=alpha2,betaneg=beta2,shift=shift))
   }
}


# require: mean mu, variance matrix sigma, similarity matrix A
# return: weight(vector), center(vector), shift(quantity)
#         and rank (of matrix W)
get_weight_center_shift = function(mu, sigma, A)
{
   # first check if A is singular
   k1 = nrow(A)                    # number of rows of A
   #temp = eigen(A,symmetric=T)
   #index = NULL
   #for(i in 1:k1)
   #{
   #   # record the indices with eigen value being 0
   #   if (abs(temp$values[i]) < MIN) index = c(index,i)
   #}
   #if (length(index) > 0)
   #{
   #   A = diag(temp$values[-index])
   #   transform = temp$vectors[,-index]
   #   if (sum(abs(mu)) > 0) mu = t(transform)%*%mu
   #   sigma = t(transform)%*%sigma%*%transform
   #}

   k2 = nrow(sigma)                 # number of rows of sigma
   temp = eigen(sigma,symmetric=T)
   r = 0
   while (r < k2)
   {
      if (temp$values[r + 1] > MIN)
         r = r + 1
      else
         break
   }
   zero = rep(0,length = k1-r)
   U = temp$vectors[,1:r]           # eigenvector of variance matrix sigma
   lambda = sqrt(temp$values[1:r])  # squareroot of eigenvalues of sigma

   W = diag(lambda,r)%*%t(U)%*%A%*%(U)%*%diag(lambda,r)
   temp = eigen(W,symmetric=T)
   V = temp$vectors
   weight = temp$values
   center = rep(0,r)
   shift = 0
   if (sum(abs(mu)) > 0)
   {
      center = diag(1/weight,r)%*%t(V)%*%diag(lambda,r)%*%t(U)%*%A%*%mu
      shift = as.numeric(t(mu)%*%A%*%mu-sum(center^2*weight))
   }
   return(list(weight = c(weight,zero), center = c(center,zero), shift = shift, rank = r))
}

# get the noncentral chi-square approximation (used in 4-cum approximation)
# require: mean mu, variance matrix Sigma and similarity matrix A
# return: df (degrees of freedom), delta(center), alpha(scale) and beta(shift)
#         of the approximated chi-square random variable
get_noncentral_chi = function(mu,Sigma, A)
{
   summu = sum(abs(mu))
   asig=A%*%Sigma;                 # A Sigma_n
   ka=matrix(0,4,2);               # Kappa_{k, 1 or 2}, k=1,2,3,4
   ka[1,1]=sum(diag(asig));        # Kappa_{1,1}
   if(summu > 0)
      ka[1,2]=t(mu) %*% A %*% mu;  # Kappa_{1,2}
   kk=rep(0,4);                    # Kappa_k, k=1,2,3,4
   kk[1] = ka[1,1]+ka[1,2]
   atemp=asig;
   for(k in 2:4)
   {                               # calculate Kappa's
      if(summu > 0)                # Kappa_{k,2}
         ka[k,2]=t(mu) %*% atemp %*% A %*% mu
      atemp=atemp %*% asig
      ka[k,1]=sum(diag(atemp))     # Kappa_{k,1}
      kk[k]=2^(k-1)*factorial(k-1)*(ka[k,1]+k*ka[k,2])
   }
   sigmaq=sqrt(kk[2]);             # sigma_Q
   s1=kk[3]/(kk[2]^(3/2))/sqrt(8); # s_1
   s2=kk[4]/(kk[2]^2)/12;          # s_2
   if(s1^2 > s2)
   {
      a=1/(s1-sqrt(s1^2-s2));       # a
      delta=s1*a^3-a^2;             # delta
      l=a^2-2*delta;                # l, degree freedom
   }  else
   {
      delta=0;                      # delta
      l=1/(s1^2);                   # l, degree freedom
   }
   mux=l+delta                     # mu_x
   sigmax=sqrt(2*(l+2*delta));     # sigma_x
   alpha = sigmax/sigmaq
   beta = mux - alpha*kk[1]        # Tstat also equal to alpha*d+beta
   return(list(df=l,delta=delta,alpha=alpha,beta=beta))
}

# each row is a haplotype
# method: 1 = matching measure
#         2 = length measure
#         3 = counting measure
find_similarity = function(hap, method=3, missing=".")
{
   nh = nrow(hap)             # number of haplotypes
   ns = ncol(hap)             # number of sites
   s = matrix(1,nh,nh)
   if (method == 1)           # matching measure
   {
      for(i in 1:(nh-1))
      {
         for(j in (i+1):nh)
         {
            k = 1
            while (k <= ns)
            {
               if (hap[i,k] == hap[j,k] || hap[i,k] == missing || hap[j,k] == missing)
                  k = k + 1
               else
                  break
            }
            if (k <= ns) s[i,j] = s[j,i] = 0
         }
      }
   }
   else if (method == 2)      # length measure
   {
      for(i in 1:(nh-1))
      {
         for(j in (i+1):nh)
         {
            llden = l = maxl = 0
            for(k in 1:ns)
            {
               if (hap[i,k] != missing && hap[j,k] != missing)
               {
                  llden = llden + 1
                  if (hap[i,k] == hap[j,k])
                  {
                     l = l + 1
                  } else
                  {
                     if (l > maxl) maxl = l
                     l = 0
                  }
               }
            }
            if (l > maxl) maxl = l
            if (llden == 0) s[i,j] = s[j,i] = 0
            else s[i,j] = s[j,i] = maxl/llden
         }
      }
   }
   else                        # counting measure
   {
      one = rep(1,length = ns)
      for(i in 1:(nh-1))
      {
         for(j in (i+1):nh)
         {
            llden = llnum = 0
            for(k in 1:ns)
            {
               if (hap[i,k] != missing && hap[j,k] != missing)
               {
                  llden = llden + 1
                  if (hap[i,k] == hap[j,k]) llnum = llnum + 1
               }
            }
            if (llden == 0) s[i,j] = s[j,i] = 0
            else s[i,j] = s[j,i] = llnum/llden
         }
      }
   }
   return (s)
}

# n is the sample size
# p: frequencies of these haplotypes
sample_by_freq = function(n, p)
{
   nhap = length(p)
   ss = sample(1:nhap,n,T,prob=p)
   phat = rep(0,nhap)
   for(k in 1:nhap) phat[k] = length(ss[ss==k])/n
   return (phat)
}

# the following are functions to obtain the density function, cumulative density function and quantiles of the difference of two chi-squares
# X1~Chisq(df1, delta1), Y1=(X1-beta1)/alpha1
# X2~Chisq(df2, delta2), Y2=(X2-beta2)/alpha2

## self-defined way to calculate density of Y1 - Y2
dchidiff = function (z, df1=1, delta1=0, alpha1=1, beta1=0, df2=1, delta2=0, alpha2=1, beta2=0)
{
   d1 = function(x) {y=alpha1*x+beta1; dchisq(y, df=df1, ncp=delta1)*alpha1;}  # pdf of Y1
   d2 = function(x) {y=alpha2*x+beta2; dchisq(y, df=df2, ncp=delta2)*alpha2;}  # pdf of Y2

   int12 = function(y) { d1(y+z)*d2(y); };
   integrate(int12, -Inf, Inf, subdivisions=1000, rel.tol =.Machine$double.eps^0.5)$value;
}
# Example: dchidiff(6, df1=3, delta1=0.5, alpha1=3, beta1=1, df2=5, delta2=1, alpha2=8, beta2=3)
# [1] 0.0007698747

## self-defined way to calculate cdf of Y1 - Y2, need package "adapt", updated on 09/23/2009
library(adapt)
pchidiff = function(z, df1=1, delta1=0, alpha1=1, beta1=0, df2=1, delta2=0, alpha2=1, beta2=0, error=1e-4, minpts = 300, eps = 1e-3)
{
   temp = Inf;
   if(df1 < 2) temp = dchisq(qchisq(error, df=df1, ncp=delta1), df=df1, ncp=delta1);
   d1 = function(x) {y=alpha1*x+beta1; ans=dchisq(y, df=df1, ncp=delta1)*alpha1; pmin(ans, temp);}  # pdf of Y1
   temp = Inf;
   if(df2 < 2) temp = dchisq(qchisq(error, df=df2, ncp=delta2), df=df2, ncp=delta2);
   d2 = function(x) {y=alpha2*x+beta2; ans=dchisq(y, df=df2, ncp=delta2)*alpha2; pmin(ans, temp);}  # pdf of Y2

   x = exp(z)/(1+exp(z));
   int12 = function(y) {
      ya <- y[1]/(1-y[1]);
      yb <- y[2]/(1-y[2]);
      d1(log(ya*yb))*d2(log(yb))/(y[1]*y[2]*(1-y[1])*(1-y[2]));
   };

   WHOLE=adapt(2, lo = c(0, 0), up = c(1, 1), functn=int12, minpts = minpts, eps = eps)$value;
   adapt(2, lo = c(0, 0), up = c(x, 1), functn=int12, minpts = minpts, eps = eps)$value/WHOLE;
}
# Example: pchidiff(-1, df1=3, delta1=0.5, alpha1=3, beta1=1, df2=5, delta2=1, alpha2=8, beta2=3)
# [1] 0.04474556 (old)  0.04451074 (new)
# Bad example: pchidiff(1, df1=2, delta1=0, alpha1=1, beta1=0, df2=1, delta2=0, alpha2=1, beta2=0)
# [1] 0.5623352 (50 secs)

## self-defined way to calculate quantile of Y1 - Y2, need package "adapt", updated on 09/23/2009
## need function pchidiff()
qchidiff = function(p, df1=1, delta1=0, alpha1=1, beta1=0, df2=1, delta2=0, alpha2=1, beta2=0, error=1e-3, minpts=300, eps=5e-3)
{
   # p = P(Y1 <= upper) = P(Y1 - Y2 < upper - Y2) <= P(Y1 - Y2 <= upper)
   upper = (qchisq(p, df=df1, ncp=delta1) - beta1)/alpha1 + beta2/alpha2

   # p = P(Y2 >= lower) = P(Y1 - Y2 <= Y1 - lower) >= P(Y1 - Y2 <= -lower)
   lower = -(qchisq(1-p, df=df2, ncp=delta2) - beta2)/alpha2 - beta1/alpha1

   ftemp = function(x)
      {pchidiff(x,df1=df1,delta1=delta1,alpha1=alpha1,beta1=beta1,df2=df2,delta2=delta2,alpha2=alpha2,beta2=beta2,error=error,minpts=minpts,eps=eps) - p}
   uniroot(ftemp, c(lower, upper), tol = p/100)$root
}
# Example: qchidiff(0.001, df1=3, delta1=0.5, alpha1=3, beta1=1, df2=5, delta2=1, alpha2=8, beta2=3)
# [1] -2.345771 (old) -2.346533 (new, 38 secs)
# Bad Example: qchidiff(0.001, df1=1, delta1=0.5, alpha1=2, beta1=1, df2=3, delta2=1, alpha2=8, beta2=3)
#[1] -2.471190 (530 secs)

